/*
 *  Checker.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 *
 *  Valentin Tablan, 03/03/2000
 *
 *  $Id: Checker.java,v 1.9 2000/11/08 16:34:56 hamish Exp $
 */

package gate.db;
import java.sql.*;
import java.util.*;
import gate.*;
import gate.db.Database.*;
import gate.util.*;

/**
  * This class provides some static methods needed by other classes for
  * database check/setup-ing.
  * It also  holds the database structure needed by Gate2 in order to be
  * able to work with a given database.
  */
public class Checker{

  /** Debug flag */
  private static final boolean DEBUG = false;

  /** This method attempts to create the structure defined by ddl.
    * Before trying to create each table it will check if such a table doesn't
    * already exist. If it finds a table with the given name, it will check if
    * the structure of the table complies with the requested structure.
    * If it cannot find a table with the given name, it will create a table with
    * the requested name and structure.
    * @param ddl An array of table definitions (using gate.db.Database.Table).
    * This parameter is actually an array of objects of type gate.db.Table.
    * An example of how to statically define such an array can be found at
    * the bottom of this file. For aditional information regarding working with
    * Table objects please read the documentation from gate.db.Table.
    * @return true if the structure of the database complies with the request or
    * if the structure of the database has been successfuly altered so it now
    * complies with the request.
    * false if the database could not be altered such that it complies with the
    * request (e.g. one of the tables exists and has a different structure).
    * Note that this method will fail (return false) only if the structure found
    * in the database cannot be converted so it complies with the request. It
    * will not fail for example if a column in a table has a size larger than
    * the requested one but it will fail if the size found in the database is
    * smaller than the requested one.
    */
  public static boolean setupDatabase(Table[] ddl, DataStore ds) {
    boolean res = true;
    Database db;
    DatabaseMetaData md;
    String currentTable;

    if ( ds instanceof Database ) {
      db = (Database)ds;
      Map dictionary = db.getDictionary();
      Connection con = ((Database)db).getConnection();
      errMsg="";
      ResultSet rsTables, rsColumns, rsPK;

      try {
        md = con.getMetaData();
        for(int i = 0; i < ddl.length; i++){
          //foreach table description
          currentTable = ddl[i].name;
          rsTables = md.getTables(db.getCatalog(),db.getSchema(),
                                  currentTable,null);
          if(rsTables.next()){
            //yes, we found a table with the right name
            //Out.println("\n\nTable:"+ddl[i].name);
            Collection metadataFields = new Vector();
            Collection ddlFields = new Vector();

            for(int j=0; j < ddl[i].fields.length; j++)
              ddlFields.add(ddl[i].fields[j]);

            //let's read the primary keys for this table
            rsPK = md.getPrimaryKeys(db.getCatalog(),db.getSchema(),
                                     currentTable);

            Collection primaryKeys = new Vector();

            while(rsPK.next()){
              primaryKeys.add(rsPK.getString("COLUMN_NAME"));
            };

            //let's read the columns for this table
            rsColumns = md.getColumns(db.getCatalog(),db.getSchema(),
                                      currentTable,"%");

            while(rsColumns.next()){
              //foreach column description
              String colName = rsColumns.getString("COLUMN_NAME");
              short colType = rsColumns.getShort("DATA_TYPE");

              String colTypeStr="N/A";
              switch (colType){
                case Types.ARRAY:{
                  colTypeStr = "ARRAY";
                  break;
                }
                case Types.BIGINT:{
                  colTypeStr = "BIGINT";
                  break;
                }
                case Types.BINARY:{
                  colTypeStr = "BINARY";
                  break;
                }
                case Types.BIT:{
                  colTypeStr = "BIT";
                  break;
                }
                case Types.BLOB:{
                  colTypeStr = "BLOB";
                  break;
                }
                case Types.CHAR:{
                  colTypeStr = "CHAR";
                  break;
                }
                case Types.CLOB:{
                  colTypeStr = "CLOB";
                  break;
                }
                case Types.DATE:{
                  colTypeStr = "DATE";
                  break;
                }
                case Types.DECIMAL:{
                  colTypeStr = "DECIMAL";
                  break;
                }
                case Types.DISTINCT:{
                  colTypeStr = "DISTINCT";
                  break;
                }
                case Types.DOUBLE:{
                  colTypeStr = "DOUBLE";
                  break;
                }
                case Types.FLOAT:{
                  colTypeStr = "FLOAT";
                  break;
                }
                case Types.INTEGER:{
                  colTypeStr = "INTEGER";
                  break;
                }
                case Types.JAVA_OBJECT:{
                  colTypeStr = "JAVA_OBJECT";
                  break;
                }
                case Types.LONGVARBINARY:{
                  colTypeStr = "LONGVARBINARY";
                  break;
                }
                case Types.LONGVARCHAR:{
                  colTypeStr = "LONGVARCHAR";
                  break;
                }
                case Types.NULL:{
                  colTypeStr = "NULL";
                  break;
                }
                case Types.NUMERIC:{
                  colTypeStr = "NUMERIC";
                  break;
                }
                case Types.OTHER:{
                  colTypeStr = "OTHER";
                  break;
                }
                case Types.REAL:{
                  colTypeStr = "REAL";
                  break;
                }
                case Types.REF:{
                  colTypeStr = "REF";
                  break;
                }
                case Types.SMALLINT:{
                  colTypeStr = "SMALLINT";
                  break;
                }
                case Types.STRUCT:{
                  colTypeStr = "STRUCT";
                  break;
                }
                case Types.TIME:{
                  colTypeStr = "TIME";
                  break;
                }
                case Types.TIMESTAMP:{
                  colTypeStr = "TIMESTAMP";
                  break;
                }
                case Types.TINYINT:{
                  colTypeStr = "TINYINT";
                  break;
                }
                case Types.VARBINARY:{
                  colTypeStr = "VARBINARY";
                  break;
                }
                case Types.VARCHAR:{
                  colTypeStr = "VARCHAR";
                  break;
                }
              }
              int colSize = rsColumns.getInt("COLUMN_SIZE");
              int colScale = rsColumns.getInt("DECIMAL_DIGITS");
              boolean colNotNull =
                rsColumns.getString("IS_NULLABLE").equalsIgnoreCase("NO");
              boolean colPK = primaryKeys.contains(colName);
              Field columnDesc = new Field(colName, colTypeStr, colSize,
                                           colScale, colNotNull,colPK);
              //Out.println(columnDesc.toString());
              metadataFields.add(columnDesc);
            }//while(rsColumns.next())
            //Out.println("\n----------------------\n");
            //now we have a collection with all the columns described in
            //metadata (metadataFields).
            if(!metadataFields.containsAll(ddlFields)){

              //there are some required columns that were NOT found in the
              //databse metadata
              res = false;

              //let's see what columns are missing
              //removeAll is not implemented
              ddlFields.removeAll(metadataFields);

              //we should actually check whether we can rectify this problem
              errMsg += "\nThere were missing columns in "+currentTable+":\n";
              Iterator missingFields = ddlFields.iterator();

              while(missingFields.hasNext()){
                errMsg += missingFields.next().toString()+";\n";
              }

            }//if(!metadataFields.containsAll(ddlFields))

          } else {
            //No table with the name &currentTable
            //let's create the missing table
            String createStmt = "CREATE TABLE "+currentTable+" (";
            String primaryKeys = null;
            String type;
            int colCnt = ddl[i].fields.length;

            for(int j=0; j < colCnt-1; j++) {
              //foreach column in ddl
              Field currentCol = ddl[i].fields[j];
              createStmt += currentCol.name + " ";
              type = currentCol.type;
              //Out.print(type+"<>");

              if(dictionary.containsKey(type))
                type = (String)dictionary.get(type);
              //Out.println(type);
              createStmt += type;

              if(currentCol.size != -1) {
                createStmt += "("+currentCol.size;
                if(currentCol.scale != -1) createStmt += ","+currentCol.scale;
                createStmt +=") ";
              } else createStmt += " ";

              if(currentCol.notNull) createStmt += "NOT NULL ";

              if(currentCol.primaryKey) {
                if(primaryKeys == null){
                  //first PK member
                  primaryKeys = currentCol.name;
                } else {
                  //not first PK member
                  primaryKeys += ", "+currentCol.name;
                }
              }

              //createStmt += "PRIMARY KEY ";
              createStmt += ", ";

            }//for(int j=0; j<ddl[i].fields.lenght; j++)

            //last column (no final comma)
            Field currentCol = ddl[i].fields[colCnt-1];

            createStmt += currentCol.name + " ";
            type = currentCol.type;

            //Out.print(type+"<>");
            if(dictionary.containsKey(type))
              type = (String)dictionary.get(type);

            //Out.println(type);
            createStmt += type;
            if(currentCol.size != -1){
              createStmt += "("+currentCol.size;
              if(currentCol.scale != -1) createStmt += ","+currentCol.scale;
                createStmt +=") ";
            } else createStmt += " ";

            if(currentCol.notNull) createStmt += "NOT NULL ";
              if(currentCol.primaryKey) {
                if(primaryKeys == null) {
                  //first PK member
                  primaryKeys = currentCol.name;
                } else {
                  //not first PK member
                  primaryKeys += ", "+currentCol.name;
                }
              }
            //final ")"

            if(primaryKeys !=null) createStmt +=",PRIMARY KEY (" +
                                                primaryKeys+") ";

            createStmt +=")";
            Statement stmtCreate = con.createStatement();
            //Out.println(createStmt);
            stmtCreate.executeUpdate(createStmt);
            errMsg += "No "+currentTable+" table...created!\n";

            //execute all other statements related to this new table
            if(ddl[i].statements != null) {
              Statement stmtOthers = con.createStatement();
              for(int j = 0; j< ddl[i].statements.length; j++){
                stmtOthers.addBatch(ddl[i].statements[j]);
              } // forall statements
              stmtOthers.executeBatch();
            } // if (ddl[i].statements != null)

          } // end create missing table
        } // end foreach table description
        // save all the changes
        con.commit();

        try {
          db.takeBackConnection(con);
        } catch (GateException ge) {
          ge.printStackTrace(Err.getPrintWriter());
        }

      } catch(java.sql.SQLException e) {

        try {
          con.rollback();
        } catch (SQLException sqle) {
          throw( new RuntimeException("Exception before commit..."+
                                      "while trying to rollback caught:\n"+
                                      e.toString()));
        };
        throw( new RuntimeException(e.toString()));
      }
    } else {
      errMsg = "Unknown datastore type: "+ds.getClass();
      //the provided datastore is not a known datastore
      return false;
    }

    if(errMsg.equals("")) errMsg = "DataStore compliant!";

    return res;
  }

  public static boolean setupDS(DataStore ds) {
    return setupDatabase(DDL, ds);
  } // setupDS


  /** The error message for the last fail of setupDB execution */
  public static String errMsg="Default";

  /** This represents the database structure required by Gate2 to run.
    * These are tables that need to exist regardless of the contents of the
    * database (that is they need to be there even in a database that contains
    * no resources)
    */
  // Gate2 requirements regarding database structure
  protected static Table[] DDL;
  static {
    // static initializer used for building the DDL structure.
    DDL = new Table[4];
    DDL[0] = new Table("FOO",
               new Field[]{
                  new Field("NAME", "VARCHAR", 20, -1, true, true),
                  new Field("TEMPBLOB", "BLOB", -1, -1, false, false)
               },
               null
             );
    DDL[1] = new Table("NEXTLRID",
               new Field[]{
                  new Field("NEXTID", "INTEGER", -1, -1, false, false),
               },
               null
             );
    DDL[2] = new Table("LR",
               new Field[]{
                  new Field("ID", "INTEGER", -1, -1, true, true),
                  new Field("TYPE", "VARCHAR", 20, -1, false, false),
                  new Field("NAME", "VARCHAR", 50, -1, false, false),
               },
               null
             );
    DDL[3] = new Table("FEATURES",
               new Field[]{
                  new Field("OWNERID", "INTEGER", -1, -1, true, true),
                  new Field("ATTRIBUTE", "VARCHAR", 20, -1, true, true),
                  new Field("VALUETYPE", "VARCHAR", 10, -1, true, false),
                  new Field("INTVALUE", "INTEGER", -1, -1, false, false),
                  new Field("LONGVALUE", "LONG", -1, -1, false, false),
                  new Field("DOUBLEVALUE", "DOUBLE", -1, -1, false, false),
                  new Field("STRINGVALUE", "VARCHAR", 100, -1, false, false),
                  new Field("COMPLEXVALUE", "BLOB", -1, -1, false, false),
               },
               null
             );

  }
  
} // class Checker
