/*
 *  CorpusWrapper.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 *
 *  Valentin Tablan 21 Feb 2000
 *
 *  $Id: CorpusWrapper.java,v 1.20 2000/11/13 17:55:17 valyt Exp $
 */


package gate.db;

import gate.*;
import gate.util.*;
import java.sql.*;
import gate.corpora.*;
import java.util.*;
import java.net.URL;
import java.io.*;
import gate.db.Database.*;

/** This class extends CorpusImpl adding database persistence via Jdbc.
  */
public class CorpusWrapper extends CorpusImpl implements LRDBWrapper{

  /** Debug flag */
  private static final boolean DEBUG = false;

  /** Constructs a CorpusWrapper object retrieving the information from the
    * DataStore provided as parameter. If there is a corpus by that name in the
    * database it will be retrieved and used for the creation of the new object;
    * if not, a new corpus will be created.
    * @param db An object of type gate.db.DataStore which this object will use
    * to retrieve/store data from/to.
    * @param name The name of the new corpus that will be retrieved from the
    * database or created and stored in the DataStoreImpl.
  */
  public CorpusWrapper(DataStore db, String name) throws IOException {
    this.name = name;
    // if(features == null) features = Factory.newFeatureMap();
    if(db instanceof Database){
      database = (Database)db;
    }else{
      throw new RuntimeException(
        "I cannot use this datastore for storing myself!\n"+
        "You should check/setup the datstore object calling setupDS on the "+
        "transient corpus before attempting to create a persistent corpus!");
    }
    DBInit();
  }

  /**
   * Creates a persistent corpus with the given name and features.
   * @param db the datastore this corpus is to be read/saved from/to.
   * @param name the name of the new corpus.
   * @param features the features of the corpus.
   */
  public CorpusWrapper(DataStore db, String name, FeatureMap features)
         throws IOException
  {
    this.name = name;
    this.features = features;
    if(db instanceof Database){
      database = (Database)db;
    } else {
      throw new RuntimeException(
       "I cannot use this datastore for storing myself!\n"+
       "You should check/setup the datastore object calling \"setupDS\" on "+
       "the transient corpus before attempting to create a persistent corpus!");
    }
    DBInit();
  }

  /**
   * Creates a persistent corpus starting from a transient one.
   * Note that this method can be used to create a persistent copy
   * of another persistent corpus located in a different database.
   *
   * @param db the datstore used for saving the new persistent corpus.
   * @param corpus the transient corpus.
   */
  public CorpusWrapper(DataStore db, CorpusImpl corpus)
         throws IOException
  {
    name = corpus.getName();
    features = corpus.getFeatures();
    Iterator docsIter = corpus.iterator();

    while(docsIter.hasNext()){
      this.add(docsIter.next());
    }

    if(db instanceof Database){
      database = (Database)db;
    }else{
      throw new RuntimeException(
        "I cannot use this datastore for storing myself!\n"+
        "You should check/setup the datastore object calling \"setupDS\" on "+
        "the transient corpus before attempting to create a persistent corpus!"
        );
    }
    DBInit();
  }

  /**
    * This method takes care of reading the a corpus from the databse or
    * creating a new one starting from the given name.
    * This method is actually the persistent part of any of the constructors.
    *
    * First this method will try to read from the database the corpus with the
    * given name. If it cannot find such a corpus, it will create a new one
    * using the values from the transient members of this object, memebers that
    * have been set by the constructor before calling DBInit.
    */
  private void DBInit() throws IOException{
    con = database.getConnection();
    try{
      beginTransaction();
      Iterator pairsIter;
      //is there a corpus with the right name?
      stmt = con.createStatement();

      //ResultSet.TYPE_FORWARD_ONLY,
      //	                         ResultSet.CONCUR_UPDATABLE);
      rsLR = stmt.executeQuery("SELECT ID, TYPE, NAME "+
                               "FROM LR "+
                               "WHERE NAME='"+name+"' "+
                               "AND TYPE='CORPUS'");

      if(rsLR.next()){
        //There is a corpus with the right name...
        //let's find out its ID
        id = rsLR.getLong("ID");

        //let's read the features
        //CAUTION!
        //If this persistent object is build starting from a transient corpus
        //which already has some features and if the database contains already
        //a persistent copy of the same object (an object with the same name)
        //the features from the database will be ADDED to those in the transient
        //object's feature map.
        //By default we consider that the values form the database are older,
        //so if the same attribute has different values in the transient object
        //and in the persistent object, the tansient values will be kept.
        FeatureMap persistentFM = database.readFeaturesOf(id, con);
        if(features != null) persistentFM.putAll(features);
        features = persistentFM;

        //now we have the correct values in the transient Feature Map
        //let's update the database too
        database.deleteFeaturesOf(id, con);
        database.writeFeaturesOf(id,features, con);
        /*
        //Oracle result sets are not sensitive, we need to repopulate it
        //Actually I'm not really sure we need this result set but I'll keep it
        //for now.
        rsFeatures = stmt.executeQuery("SELECT OWNERID, ATTRIBUTE, VALUETYPE, "+
                                       "NUMBERVALUE, STRINGVALUE,COMPLEXVALUE "+
                                       "FROM FEATURES "+
                                       "WHERE OWNERID="+id);
        */
        //Let's read the documents...
        //Delete all the documents already in this corpus
        //It should be empty but let's make sure about that
        this.clear();
        rsCorpusDocument = stmt.executeQuery(
            "SELECT CORPUS.ID AS CID, DOCID, "+
            "URL, DOCSTART, DOCEND "+
            "FROM CORPUS, DOCUMENT "+
            "WHERE CORPUS.ID = "+id+" "+
            "AND DOCID = DOCUMENT.ID");
        String url;
        long docStart, docEnd, docId;

        while(rsCorpusDocument.next()){
          docId= rsCorpusDocument.getLong("DOCID");
          url = rsCorpusDocument.getString("URL");
          docStart = rsCorpusDocument.getLong("DOCSTART");
          docEnd= rsCorpusDocument.getLong("DOCEND");
          URL docUrl=null;

          try {
            docUrl=new URL(url);
          } catch(java.net.MalformedURLException mue) {
            mue.printStackTrace(Err.getPrintWriter());
          };

          try{
            DocumentWrapper doc = new DocumentWrapper(con, database,docId);
            this.add(doc);
          } catch(NoSuchObjectException nsoe) {
            nsoe.printStackTrace(Err.getPrintWriter());
          }

        } // while rsCorpusDocument.next
        isNew = false;
      } else {
        //No corpus with that name -> We have to create a new one.
        //let's get a new id for this corpus.
        id = Tools.genTime().intValue();
        stmt.executeUpdate("INSERT INTO LR (ID, TYPE, NAME) "+
                           "VALUES ("+id+", 'CORPUS', '"+name+"')");

        //let's write the features in the database
        if(features != null) {
          database.writeFeaturesOf(id, features, con);
        };//if(features != null)

        //let's write the documents
        if(!this.isEmpty()){
          Iterator docsIter = this.iterator();
          DocumentImpl currentDoc;
          PreparedStatement psCorpus = con.prepareStatement(
                            "INSERT INTO CORPUS (ID, DOCID) "+
                            "VALUES("+id+",?)");

          while(docsIter.hasNext()){
            currentDoc = (DocumentImpl)docsIter.next();
            DocumentWrapper persistentDoc =
                            new DocumentWrapper(con, database, currentDoc);
            psCorpus.setLong(1, persistentDoc.getId());
            psCorpus.executeUpdate();
          }
        }
        isNew = true;
      }
    } catch(SQLException sqle) {
      sqle.printStackTrace(Err.getPrintWriter());
    }
  } // we have read or created the new corpus

  /** Rolls back all the actions performed since the last beginTransaction call
    */
  public void rollback() throws SQLException{
    con.rollback();
  } // rollback

  /** Starts a transaction. All database actions should be enclosed between a
    * beginTransaction() and a commit()/rollback() call.
    */
  public void beginTransaction() throws SQLException{
    con.setAutoCommit(false);
  } // beginTransaction

  /** Commits the reads/writes */
  public void commit() throws SQLException{
    //let's commit the transaction
    con.commit();
    con.setAutoCommit(true);
  } // commit

  /** Disconnects from the Datastore object, that is returns the connection */
  public void disconnect() throws GateException{
    database.takeBackConnection(con);
  } // disconnect

  /** Returns the datastore object used for the persistent store of this cropus
    */
  public DataStore getDataStore(){
    return database;
  } // getDataStore

  /** This method can be used to check whether the constructor read an old
    * corpus previously stored in in the database or it created a new corpus
    * starting from transient values.
    */
  public boolean isNew(){
    return isNew;
  } // isNew

  /** This method checks whether the provided datastore can be used for storing
    * persistent objects of this type(Corpus). If necessary, this method will
    * try to adapt the given datastore so it complies with the requirements
    * needed for storing corpora.
    * @return true if the datastore is corpus-compliant or if the datastore has
    * been successfully adapted;
    * false otherwise.
    */
  public static boolean setupDatabase(DataStore ds){
    return Checker.setupDatabase(DDL,ds);
  }

  /** @return the id used for this corpus in the database.
    */
  public long getId(){
    return id;
  } // getId

  protected long id;

  protected Database database;

  protected Connection con;

  protected Statement stmt;

  protected ResultSet rsLR;

  protected ResultSet rsCorpusDocument;

  protected ResultSet rsFeatures;

  protected boolean isNew;

  /** The definition for database requirements of persistent Corpora.
    * THis array contains (static)objects of type gate.db.Table which are used
    * to define the table structure of the database.
    * A Table object takes the following as paramateres at construction time:
    *   -the name of the table;
    *   -an array of Field objects (used to define the table columns);
    *   -an array of String objects (used to define any suplemetary SQL
    *      statements
    * needed to initialize the database (e.g. an INSERT statement which adds
    * some initial values in the table or an ALTER TABLE statement, etc)).
    */
  protected static Table[] DDL;
  static{
    //static initializer used for building the DDL structure.
    DDL = new Table[2];
    DDL[0] = new Table("CORPUS",
               new Field[]{
                  new Field("ID", "INTEGER", -1, -1, true, true),
                  new Field("DOCID", "INTEGER", -1, -1, true, true)
               },
               null
             );
    DDL[1] = new Table("DOCUMENT",
               new Field[]{
                  new Field("ID", "INTEGER", -1, -1, true, true),
                  new Field("URL", "VARCHAR", 200, -1, true, false),
                  new Field("DOCSTART", "NUMERIC", 20, -1, false, false),
                  new Field("DOCEND", "NUMERIC", 20, -1, false, false)
               },
               null
             );
  }

} // class CorpusWrapper
