/*
 *  Database.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 *
 *  Valentin Tablan 21 Feb 2000
 *
 *  $Id: Database.java,v 1.15 2001/01/22 18:59:33 hamish Exp $
 */
package gate.db;

import java.sql.*;
import gate.*;
import gate.util.*;
import java.util.*;
import java.io.*;

/**
  * Objects of type Database are used to connect to a database. They hold all
  * the necessary information for accessing the database and provide services
  * to client objects (i.e. persistent language resources).
  */
public class Database
extends gate.persist.SerialDataStore // quick and dirty way of implementing
                                     // the DS API
{
    /** Debug flag */
    private static final boolean DEBUG = false;

    /** Constructs an object of type database.
     * @param driver an object of type java.sql.Driver. The driver used for
     * JDBC access to the database.
     * @param url the JDBC url that identifies the database this object
     * should connect to.
     * @param user The user name for accessing the database.
     * @param password The passwordfor accessing the database.
     * @param catalog The catalog where the tables will be found.
     * @param schema The schema where the tables will be found.
     */
    public Database(Driver driver, String url, String user, String password,
                  String catalog, String schema){
    this.driver = driver;
    this.url = url;
    this.user = user;
    this.password = password;
    this.catalog = catalog;
    this.schema = schema;
  }

  /**
    * This method needs to be called before any access is provided to the
    * database. This is why this method gets called the first time
    * a connection is requested using "getConnection".
    * Note that the Database needs to have a valid format. See setupDS methods
    * on gate.db.Checker, CorpusWrapper, DocumentWrapper.
    */
  protected void connect() {
    //load the driver
    try {
      //better than Class.forname() which works only in Sun's JDK
      //who knows, maybe somebody uses Microsoft's SDK for some (strange) reason
      DriverManager.registerDriver(driver);
    } catch(java.sql.SQLException sqlx) {
      throw( new RuntimeException(driver+" could not be located!"));
    }

    //connect to the Database
    Connection con=null;
    try{
      con = DriverManager.getConnection(url, user, password);
      // con.setTransactionIsolation(TRANSACTION_LEVEL);
    }catch(java.sql.SQLException e){
      throw( new RuntimeException(e.toString()));
    }
    connections.add(con);
    availableConnections.add(con);
  } // connect

  /**
    * This method is used by the clients of this Database object in order
    * to gain access to the database.
    * @return an object of type java.sql.Connection providing access to the
    * Database that this Database object is associated to.
    * If this object is not yet connected to the physical database, the first
    * call of getConnection will also make the Jdbc connection to the database.
    * All connections obtained via this method should be returned using
    * takeBackConnection().
    */
  public synchronized Connection getConnection(){
    //get a connection
    Connection con=null;
    if(!isConnected()) connect();

    try {
      if(availableConnections.isEmpty()){//no more connections -> make a new one
          con = DriverManager.getConnection(url, user, password);
          connections.add(con);
      } else { //take first available connection
        con=(Connection)(availableConnections.remove(0));
      }
    } catch(java.sql.SQLException e) {
      throw( new RuntimeException(e.toString()));
    };
    return con;
  } // getConnection

  /**
    * Takes back a previously supplied connection and makes it available
    * for other clients.
    * @param con the connection being returned.
    * If the connection was not provided by this Database object, an exception
    * will be raised.
    */
  public void takeBackConnection(Connection con)
              throws GateException
  {
    if(connections.contains(con)){//it is one of my connections
      if(availableConnections.contains(con)) {//already released
        throw new GateException(
                  "This connection has already been returned!");
      } else {//all OK, committing
        //adds to the end of the list because is a LinkedList
        availableConnections.add(con);
      }
    } else {//"You haven't got this connection from me!"
      throw new GateException(
                "This connection was not provided by this Database!");
    }
  } // takeBackConnection


  /**
    * Sets the features of this datastore.
    *
    * @param features a value of type 'FeatureMap'
    */
  public void setFeatures(FeatureMap features){
    this.features = features;
  } // setFeatures

  /**
    * Gets the features of this datastore.
    *
    * @return a value of type 'FeatureMap'
    */
  public FeatureMap getFeatures(){
    return features;
  } // getFeatures

  /**
    * Used to find out if this Database object is actually connected to a
    * database.
    *
    * @return true if it is connected, false otherwise.
    */
  public boolean isConnected() {
    return !connections.isEmpty();
  } // isConnected

  /**
    * Frees the allocated resources (closes all the open connections).
    *
    * @exception OpenTransactionsException if there are any connections
    * that haven't been returned by the clients of this database.
    */
  public void release() throws OpenTransactionsException{
    if(connections.size() == availableConnections.size()){
      //all transactions are free
      Iterator connIter = connections.iterator();
      try {
        while(connIter.hasNext()) {
          ((Connection)(connIter.next())).close();
        }
      } catch(java.sql.SQLException e) {
        throw( new RuntimeException(e.toString()));
      };
    } else {//there are still some active transactions
      throw new OpenTransactionsException(
                "There are still some active transactions!");
    }
  } // release


  /**
    * Reads a Object from a BLOB on the current row in a result set.
    * Note that the Blob must contain a java Object saved using serialisation.
    * The implemetation at this level coresponds to a generic Jdbc source.
    * More efficient implemetations for this action can be found in classes
    * derived from this class taht are dedicated to sepecific types of Jdbc
    * sources (e.g. gate.db.OracledataStore).
    * @param rs a result set positioned on the row from which the BLOB is to
    * be read
    * @param columnName the name of the column containing the BLOB on the
    * current row.
    * @return an object containing the blob.
    * @exception SQLException if an error occurs.
    */
  public Object readObjectBlob(ResultSet rs, Connection con, String columnName)
                throws SQLException, IOException
  {
    return null;
  } // readObjectBlob


  /**
    * Acts as a setBlob for the provided prepared statement on the position
    * indicated by place.
    * @param ps A prepared statement where the blob will be set.
    * @param place The position on the prepared statement where the blob
    * will be set.
    * @param value The object that will be serialised and converted into a blob.
    */
  public void putObject(PreparedStatement ps, int place, Object value)
    throws SQLException, IOException
  {
    ObjectWriter ow = new ObjectWriter(value);
    ps.setBinaryStream(place, ow.getInputStream(), ow.getSize());
    ow.start();
  } // putObject


  /**
    * This method reads (from the Gate2 database) the features of
    * the object bearing the given ID.
    *
    * @param ownerId a value of type 'long' representing the id of
    * the object for which the features are to be retrieved.
    * @param con a value of type 'Connection' which should be connceted
    * to the database from which the features are to be read.
    * @return a value of type 'FeatureMap'
    * @exception SQLException if an error occurs
    */
  public FeatureMap readFeaturesOf(long ownerId, Connection con)
    throws SQLException, IOException
  {
    Statement stmt = con.createStatement();
    //ResultSet.TYPE_SCROLL_SENSITIVE,
    //		                                 ResultSet.CONCUR_UPDATABLE);
    ResultSet rsFeatures;
    rsFeatures = stmt.executeQuery("SELECT OWNERID, ATTRIBUTE, VALUETYPE, "+
                                   "INTVALUE, LONGVALUE, DOUBLEVALUE,  "+
                                   "STRINGVALUE, COMPLEXVALUE "+
                                   "FROM FEATURES "+
                                   "WHERE OWNERID="+ownerId);
    FeatureMap persistentFM = new SimpleFeatureMapImpl();
    while(rsFeatures.next()){
      String attName = rsFeatures.getString("ATTRIBUTE");
      String valueType = rsFeatures.getString("VALUETYPE");
      Object value=null;

      if(valueType.equalsIgnoreCase("INT")){
        value  = new Integer(rsFeatures.getInt("INTVALUE"));
      }else if(valueType.equalsIgnoreCase("LONG")){
        value  = new Long(rsFeatures.getLong("LONGVALUE"));
      }else if(valueType.equalsIgnoreCase("DOUBLE")){
        value  = new Double(rsFeatures.getDouble("DOUBLEVALUE"));
      }else if(valueType.equalsIgnoreCase("STRING")){
        value  = rsFeatures.getString("STRINGVALUE");
      }else if(valueType.equalsIgnoreCase("BLOB")){
        value  = readObjectBlob(rsFeatures, con, "COMPLEXVALUE");
      }
      persistentFM.put(attName,value);
    } // while(rsFeatures.next())
    return persistentFM;
  } // readFeaturesOf

  /**
    * This method writes the given FeatureMap as features belonging
    * to the object with the given offset.
    *
    * @param ownerId a value of type 'long' representing the offset of the
    * owner of the feature map.
    * @param featuresToWrite the FeatureMap to be written.
    * @param con A connection to the Gate2 database.
    * @exception SQLException if an error occurs
    */
  public void writeFeaturesOf(long ownerId, FeatureMap featuresToWrite,
                              Connection con)
  throws SQLException, IOException
  {
    PreparedStatement psFeatures = con.prepareStatement(
                                      "INSERT INTO FEATURES (OWNERID, "+
                                      "ATTRIBUTE, VALUETYPE, INTVALUE, "+
                                      "LONGVALUE, DOUBLEVALUE, "+
                                      "STRINGVALUE, COMPLEXVALUE) "+
                                      "VALUES("+ownerId+",?,?,?,?,?,?,?)");
    Iterator pairsIter = featuresToWrite.entrySet().iterator();
    String attName;
    Object value;

    while(pairsIter.hasNext()) {
      Map.Entry entry = (Map.Entry)pairsIter.next();
      attName = (String)entry.getKey();
      value = entry.getValue();
      psFeatures.setString(1, attName);

      if(value instanceof Integer) {

        psFeatures.setString(2,"INT");
        psFeatures.setInt(3,((Integer)value).intValue());
        psFeatures.setNull(4,Types.NUMERIC);
        psFeatures.setNull(5,Types.DOUBLE);
        psFeatures.setNull(6,Types.VARCHAR);
        psFeatures.setNull(7,Types.BLOB);

      } else if(value instanceof Long) {

        psFeatures.setString(2,"LONG");
        psFeatures.setNull(3,Types.INTEGER);
        psFeatures.setLong(4,((Long)value).longValue());
        psFeatures.setNull(5,Types.DOUBLE);
        psFeatures.setNull(6,Types.VARCHAR);
        psFeatures.setNull(7,Types.BLOB);

      } else if(value instanceof Double) {

        psFeatures.setString(2,"DOUBLE");
        psFeatures.setNull(3,Types.INTEGER);
        psFeatures.setNull(4,Types.NUMERIC);
        psFeatures.setDouble(5,((Double)value).doubleValue());
        psFeatures.setNull(6,Types.VARCHAR);
        psFeatures.setNull(7,Types.BLOB);

      } else if(value instanceof String) {

        psFeatures.setString(2,"STRING");
        psFeatures.setNull(3,Types.INTEGER);
        psFeatures.setNull(4,Types.NUMERIC);
        psFeatures.setNull(5,Types.DOUBLE);
        psFeatures.setString(6,(String)value);
        psFeatures.setNull(7,Types.BLOB);

      } else {
        //everything else is a BLOB
        psFeatures.setString(2,"BLOB");
        psFeatures.setNull(3,Types.INTEGER);
        psFeatures.setNull(4,Types.NUMERIC);
        psFeatures.setNull(5,Types.DOUBLE);
        psFeatures.setNull(6,Types.VARCHAR);
        putObject(psFeatures, 7, value);
      }
      psFeatures.executeUpdate();
    };//while(pairsIter.hasNext())
  } // writeFeaturesOf

  /**
    * Deletes from the Gate2 database all the features belonging to the
    * object with the given id.
    *
    * @param ownerId a value of type 'long' representing the database ID
    * of the object for which the features are to be erased.
    * @param con a connection to the Gate2 database.
    * @exception SQLException if an error occurs
    */
  public void deleteFeaturesOf(long ownerId,Connection con)
  throws SQLException
  {
    Statement stmt = con.createStatement();
    stmt.executeUpdate("DELETE FROM FEATURES WHERE OWNERID="+ownerId);
  } // deleteFeaturesOf

  /**
    * Returns the 'catalog' associated with this database object.
    *
    * @return a value of type 'String'
    */
  public String getCatalog(){
    return catalog;
  } // getCatalog

  /**
    * Returns the 'schema' value associated with this database object.
    *
    * @return a value of type 'String'
    */
  public String getSchema(){
    return schema;
  } // getSchema

  public Map getDictionary(){
    return dictionary;
  } // getDictionary

  protected List connections  =new LinkedList();

  protected List availableConnections = new LinkedList();

  protected Driver driver;

  protected String url;

  protected String user;

  protected String password;

  protected FeatureMap features;

  protected String catalog;

  protected String schema;

  protected Map dictionary = new HashMap();

  public static int TRANSACTION_LEVEL = Connection.TRANSACTION_READ_COMMITTED;

} // Database
