/*
 *  DocumentWrapper.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 *
 *  Valentin Tablan 03 Mar 2000
 *
 *  $Id: DocumentWrapper.java,v 1.15 2000/11/13 17:55:17 valyt Exp $
 */

package gate.db;

import gate.corpora.*;
import gate.*;
import gate.util.*;
import gate.db.Database.*;

import java.sql.*;
import java.util.*;
import java.net.URL;
import java.io.*;

/** This class extends DocumentImpl adding database persistence via Jdbc.
  */
public class DocumentWrapper extends DocumentImpl implements LRDBWrapper{

  /** Debug flag */
  private static final boolean DEBUG = false;

  /** Construction from URL and offsets. Both offsets must be
    * non-null, or they are both ignored. Content is read over the
    * net between the offsets.
    */
  public DocumentWrapper(URL u, String encoding, Long start, Long end)
         throws IOException {
    // store the offsets if they're non-null
    if(start != null && end != null) {
      sourceUrlStartOffset = start;
      sourceUrlEndOffset = end;
    }

    // store the URL
    sourceUrl = u;

    // get content out of the URL
    content = new DocumentContentImpl(u, encoding, start, end);
  } // DocumentWrapper(u,encoding,start,end)

  /** Construction from URL and offsets. Both offsets must be
    * non-null, or they are both ignored. Content is read over the
    * net between the offsets.
    */
  public DocumentWrapper(DataStore ds, URL u, Long start, Long end)
          throws IOException, SQLException
  {
    this(u, null, start, end);
    if(ds instanceof Database){
      database = (Database)ds;
    }else{
      throw new RuntimeException(
        "I cannot use this datastore for storing myself!\n"+
        "You should check the datstore object calling \"setupDS\" on the "+
        "transient corpus before attempting to create a persistent corpus!");
    }
    // denotes the fact that the id in unknown so the document will be retrieved
    // using the URL
    id = -1;
    try {
      connectToDb();
      DBInit();
    } catch(NoSuchObjectException nsoe) {
      // This kind of exception should not be thrown here
      // Print it out anyway in case something is wrong
      nsoe.printStackTrace(Err.getPrintWriter());
    }
  } // DocumentImpl(u,start,end)

  /** Construction from URL; content read over the net. */
  public DocumentWrapper(DataStore ds, URL u)
         throws IOException, SQLException
  {
    this(ds, u, null, null);
  }


  /** Construction from String representing URL */
  public DocumentWrapper(DataStore ds, String urlString)
         throws IOException, SQLException
  {
    this(ds, new URL(urlString), null, null);
  } //DocumentImpl(DataStore ds, String urlString)

  /** Construction from Document. This is the recommended way for making a
    * transient document become persistent.
    */
  public DocumentWrapper(DataStore ds, DocumentImpl doc)
         throws IOException, SQLException
  {
    features = doc.getFeatures();
    namedAnnotSets = doc.getNamedAnnotationSets();
    sourceUrl = doc.getSourceUrl();
    sourceUrlStartOffset = doc.getSourceUrlStartOffset();
    sourceUrlEndOffset = doc.getSourceUrlEndOffset();
    defaultAnnots = doc.getAnnotations();

    if(ds instanceof Database){
      database = (Database)ds;
    } else {
      throw new RuntimeException(
        "I cannot use this datastore for storing myself!\n"+
        "You should check the datstore object calling \"setupDS\" on the "+
        "transient document before attempting to create a persistent one!");
    }
    //mark that the id is not known
    id = -1;

    try{
      connectToDb();
      DBInit();
    } catch(NoSuchObjectException nsoe) {
      //This kind of exception should not be thrown here
      //Print it out anyway in case something is wrong
      nsoe.printStackTrace(Err.getPrintWriter());
    }
  }


  /** Construction from Document. This is the recommended way for making a
    * transient document become persistent.
    * Unlike the other constructor that gets a document, thus one receives
    * an open connection rather than a DataStore.
    * This is useful for clients that want this document to be read/written
    * to the database inside their open transaction.
    */
  public DocumentWrapper(Connection con, Database db, DocumentImpl doc)
         throws IOException
  {
    features = doc.getFeatures();
    namedAnnotSets = doc.getNamedAnnotationSets();
    sourceUrl = doc.getSourceUrl();
    sourceUrlStartOffset = doc.getSourceUrlStartOffset();
    sourceUrlEndOffset = doc.getSourceUrlEndOffset();
    defaultAnnots = doc.getAnnotations();

    //marks that the id is not known
    id = -1;

    try {
      database = db;
      this.con = con;
      DBInit();
    } catch(NoSuchObjectException nsoe) {
      //This kind of exception should not be thrown here
      //Print it out anyway in case something is wrong
      nsoe.printStackTrace(Err.getPrintWriter());
    }
  }


  /** Construction from ID. This is the way to get the document when the Id
    * is already known.
    */
  public DocumentWrapper(DataStore ds, long id)
         throws IOException, NoSuchObjectException, SQLException
  {
    this.id = id;

    if(ds instanceof Database) {
      database = (Database)ds;
    } else {
      throw new RuntimeException(
        "I cannot use this datastore for storing myself!\n"+
        "You should check the datstore object calling \"setupDS\" on the "+
        "transient document before attempting to create a persistent one!");
    }

    connectToDb();
    DBInit();
  } // DocumentImpl(DataStore ds, String urlString)

  /** Construction from ID. This is the way to get the document when the Id
    * is already known.
    */
  public DocumentWrapper(Connection con, Database ds, long id)
         throws IOException, NoSuchObjectException, SQLException
  {
    this.id = id;
    database = ds;
    this.con = con;
    DBInit();
  } // DocumentImpl(DataStore ds, String urlString)


  /**
    * This method is actually the persistent part of any of the constructors.
    *
    * @exception NoSuchObjectException when an attempt to read an inexistant
    * object from the database(e.g. when an invalid ID is used to retrieve a
    * document).
    * @exception IOException if an error occurs while trying to read the
    * document content.
    */
  private void DBInit()
    throws IOException, NoSuchObjectException
  {

    try{

      stmt = con.createStatement();
      //ResultSet.TYPE_SCROLL_SENSITIVE,
      //	                         ResultSet.CONCUR_UPDATABLE);
      if(id != -1){

        //we have the id.
        rsDocument = stmt.executeQuery("SELECT ID, URL, DOCSTART, DOCEND "+
                                 "FROM DOCUMENT "+
                                 "WHERE ID="+id);
        if(rsDocument.next()){
          //there is such a document in the database
          //let's read it
          isNew = false;

          //read the id
          sourceUrl = new URL(rsDocument.getString("URL"));

          //read the offsets
          boolean validOffsets = true;
          long docStart = rsDocument.getLong("DOCSTART");
          if(rsDocument.wasNull()) validOffsets = false;
          long docEnd = rsDocument.getLong("DOCEND");
          if(rsDocument.wasNull()) validOffsets = false;
          if(validOffsets) {
            sourceUrlStartOffset = new Long(docStart);
            sourceUrlEndOffset = new Long(docEnd);
          }

          //build the content
          content = new DocumentContentImpl(sourceUrl,
                                            null,
                                            sourceUrlStartOffset,
                                            sourceUrlEndOffset);
          // read the features
          // CAUTION!
          // If this persistent object is build starting from a transient one
          // which already has some features and if the database contains
          // already a persistent copy of the same object (an object with the
          // same URL) the features from the database will be ADDED to those in
          // the transient object's feature map.
          // By default we consider that the values form the database are older,
          // so if the same attribute has different values in the transient
          // object and in the persistent object, the tansient values will be
          // kept.
          FeatureMap persistentFM = database.readFeaturesOf(id, con);
          if(features != null) persistentFM.putAll(features);
          features = persistentFM;
          database.deleteFeaturesOf(id,con);
          database.writeFeaturesOf(id,features,con);

          //Read annotations
          rsAnnotations = stmt.executeQuery("SELECT SETNAME,ID, TYPE, "+
                                            "STARTNODEID, ENDNODEID "+
                                            "FROM ANNOTATION "+
                                            "WHERE DOCID="+id);
          while(rsAnnotations.next()){

            String annSetName = rsAnnotations.getString("SETNAME");
            int annId = rsAnnotations.getInt("ID");
            String annType = rsAnnotations.getString("TYPE");
            Long annStart = new Long(rsAnnotations.getLong("STARTNODEID"));
            Long annEnd = new Long(rsAnnotations.getLong("ENDNODEID"));
            FeatureMap annFM = database.readFeaturesOf(annId, con);
            AnnotationSet as;

            if(annSetName == null) as = getAnnotations();
            else as = getAnnotations(annSetName);

            try {
              as.add(new Integer(annId), annStart, annEnd, annType, annFM);
            } catch(InvalidOffsetException ioe) {
              ioe.printStackTrace(Err.getPrintWriter());
            }
          } // read the annotations

          // by now we should have a transient copy of the persistent document
        } else {
          // A document with the given URL could not be found in the database
          throw new NoSuchObjectException("A document having id="+id+" could "+
                                          "not be found in the database!");
        }
      } else {
        // We only have the URL. We need to check whether it's a new or old
        // document
        rsDocument = stmt.executeQuery("SELECT ID, URL, DOCSTART, DOCEND "+
                                 "FROM DOCUMENT "+
                                 "WHERE URL='" +
                                 sourceUrl.toExternalForm()+"' ");
        if ( rsDocument.next() ) {
          // there is such a document in the database
          // let's read it
          isNew = false;

          // read the id
          id = rsDocument.getLong("ID");

          // read the offsets
          boolean validOffsets = true;
          long docStart = rsDocument.getLong("DOCSTART");
          if(rsDocument.wasNull()) validOffsets = false;
          long docEnd = rsDocument.getLong("DOCEND");
          if(rsDocument.wasNull()) validOffsets = false;
          if(validOffsets) {
            sourceUrlStartOffset = new Long(docStart);
            sourceUrlEndOffset = new Long(docEnd);
          }
          //build the content
          content = new DocumentContentImpl(sourceUrl,
                                            null,
                                            sourceUrlStartOffset,
                                            sourceUrlEndOffset);
          // read the features
          // CAUTION!
          // If this persistent object is build starting from a transient one
          // which already has some features and if the database contains
          // already a persistent copy of the same object (an object with
          // the same URL) the features from the database will be ADDED to those
          // in the transient object's feature map.
          // By default we consider that the values form the database are older,
          // so if the same attribute has different values in the transient
          // object and in the persistent object, the tansient values will
          // be kept.

          FeatureMap persistentFM = database.readFeaturesOf(id, con);
          if(features != null) persistentFM.putAll(features);
          features = persistentFM;

          // save the new constructed feature map back to the database
          database.deleteFeaturesOf(id,con);
          database.writeFeaturesOf(id,features,con);

          // Read annotations
          rsAnnotations = stmt.executeQuery("SELECT SETNAME,ID, TYPE, "+
                                            "STARTNODEID, ENDNODEID "+
                                            "FROM ANNOTATION "+
                                            "WHERE DOCID="+id);
          while(rsAnnotations.next()) {

            String annSetName = rsAnnotations.getString("SETNAME");
            int annId = rsAnnotations.getInt("ID");
            String annType = rsAnnotations.getString("TYPE");
            Long annStart = new Long(rsAnnotations.getLong("STARTNODEID"));
            Long annEnd = new Long(rsAnnotations.getLong("ENDNODEID"));
            FeatureMap annFM = database.readFeaturesOf(annId, con);
            AnnotationSet as;

            if(annSetName == null) as = getAnnotations();
            else as = getAnnotations(annSetName);
            try {
              as.add(new Integer(annId), annStart, annEnd, annType, annFM);
            } catch(InvalidOffsetException ioe) {
              ioe.printStackTrace(Err.getPrintWriter());
            }
          } // read the annotations
          // by now we should have a transient copy of the persistent document

        } else {
          // A document with the given URL could not be found in the database
          // Let's save the data we already have in the database
          isNew = true;

          // the document does not come from the database so we need to create
          // a new ID for it.
          id = Tools.genTime().intValue();

          // Write data into DOCUMENT
          PreparedStatement psDocument = con.prepareStatement(
                            "INSERT INTO DOCUMENT (ID, URL, DOCSTART, DOCEND) "+
                            "VALUES("+id+",?,?,?)");
          psDocument.setString(1, getSourceUrl().toExternalForm());

          if(sourceUrlStartOffset == null || sourceUrlEndOffset == null) {
            psDocument.setNull(2,Types.NUMERIC);
            psDocument.setNull(3,Types.NUMERIC);
          } else {
            psDocument.setLong(2, sourceUrlStartOffset.longValue());
            psDocument.setLong(3, sourceUrlEndOffset.longValue());
          }

          psDocument.execute();
          // Write features
          if(features != null) database.writeFeaturesOf(id,features,con);

          // Write data into ANNOTATION
          PreparedStatement psAnnotation = con.prepareStatement(
                            "INSERT INTO ANNOTATION (DOCID, SETNAME, ID, TYPE "+
                            "STARTNODEID, ENDNODEID) "+
                            "VALUES("+id+"?,?,?,?,?");

          PreparedStatement psNode = con.prepareStatement(
                            "INSERT INTO NODE (ID, NODEOFFSET) "+
                            "VALUES(?,?");
          Iterator annots;
          String setName;
          // read the default set first
          if(defaultAnnots != null) {
            annots = defaultAnnots.iterator();
            setName = null; //default set
            while(annots.hasNext()) {
              Annotation currentAnn = (Annotation) annots.next();
              writeAnnotation(setName,currentAnn);
            }//wrote default annotation set
          }

          // let's write the other annotation sets...
          if(namedAnnotSets !=null) {
            Iterator namedAnnSetsIter = namedAnnotSets.entrySet().iterator();

            while (namedAnnSetsIter.hasNext()) {
              // foreach named annotation set
              Map.Entry entry = (Map.Entry)namedAnnSetsIter.next();
              setName = (String)entry.getKey();
              AnnotationSet currentAnnSet = (AnnotationSet)entry.getValue();
              annots = currentAnnSet.iterator();

              while(annots.hasNext()) {
                Annotation currentAnn = (Annotation) annots.next();
                writeAnnotation(setName,currentAnn);
              } // wrote this annotation set

            } // wrote all annotation sets

          }
        }
      }// end find by URL
    }catch(SQLException sqle){
      sqle.printStackTrace(Err.getPrintWriter());
    }
  }// we have read or created the new document

  /** Gets a new connection and starts a transaction on that connection
    */
  private void connectToDb()throws SQLException{
    con = database.getConnection();
    beginTransaction();
  }

  /**
    * This method writes an annotation into the database.
    *
    * @param setName The name of the Annotation Set this annotation
    * belongs to.
    * @param ann The annotation to be written into DB.
    * @exception SQLException if an error occurs
    */
  protected void writeAnnotation(String setName, Annotation ann)
                 throws SQLException, IOException
  {
    String annType;
    long annId,sNodeId, eNodeId, sNodeOffset, eNodeOffset;
    annType = ann.getType();
    annId = ann.getId().longValue();
    sNodeId = ann.getStartNode().getId().longValue();
    sNodeOffset = ann.getStartNode().getOffset().longValue();
    eNodeId = ann.getEndNode().getId().longValue();
    eNodeOffset = ann.getEndNode().getOffset().longValue();

    if(setName == null)psAnnotation.setNull(1, Types.VARCHAR);
    else psAnnotation.setString(1, setName);

    psAnnotation.setLong(2, annId);
    psAnnotation.setString(3, annType);
    psAnnotation.setLong(4, sNodeId);
    psAnnotation.setLong(5, eNodeId);
    psAnnotation.execute();
    psNode.setLong(1,sNodeId);
    psNode.setLong(2,sNodeOffset);
    psNode.execute();
    psNode.setLong(1,eNodeId);
    psNode.setLong(2,eNodeOffset);
    psNode.execute();

    //write the features of the annotation
    FeatureMap annFM = ann.getFeatures();
    if(annFM != null) database.writeFeaturesOf(annId, annFM, con);
  }

  /** Rolls back all the actions performed since the last beginTransaction call
    */
  public void rollback() throws SQLException{
    con.rollback();
  }

  /** Starts a transaction */
  public void beginTransaction() throws SQLException{
    con.setAutoCommit(false);
  }

  /** Commits the reads/writes */
  public void commit() throws SQLException{

    // let's commit the transaction
    con.commit();
    con.setAutoCommit(true);
  }

  /** Disconnects from the Datastore object, i.e. returns the connection */
  public void disconnect() throws GateException{
    database.takeBackConnection(con);
  }

  /**
    * Returns the Datastore object associated to this persistent document.
    *
    * @return a value of type 'DataStore'
    */
  public DataStore getDataStore(){
    return database;
  }

  /**
    * Can be used to check if this document is an old one that has been
    * retrieved from the databse or is a new one that has just been saved.
    *
    * @return a value of type 'boolean'
    */
  public boolean isNew(){
    return isNew;
  }

  /**
    * Checks if the provided datastore id proper for storing persistent
    * documents of this type.
    *
    * @param ds a value of type 'DataStore'
    * @return a value of type 'boolean'
    */
  public static boolean setupDatabase(DataStore ds){
    return Checker.setupDatabase(DDL,ds);
  }

  /**
    * Gets the database ID of this persistent object.
    *
    * @return a value of type 'long'
    */
  public long getId(){
    return id;
  }

  protected long id=-1;

  // protected DocumentImpl transientDoc;
  protected Database database;

  protected Connection con;

  protected Statement stmt;

  protected ResultSet rsDocument, rsAnnotations, rsFeatures;

  protected PreparedStatement psAnnotation,psNode;

  protected boolean isNew;

  protected static Table[] DDL;

  static{
    //static initializer used for building the DDL structure.
    DDL = new Table[3];
    DDL[0] = new Table("DOCUMENT",
               new Field[]{
                  new Field("ID", "INTEGER", -1, -1, true, true),
                  new Field("URL", "VARCHAR", 200, -1, true, false),
                  new Field("DOCSTART", "NUMERIC", 20, -1, false, false),
                  new Field("DOCEND", "NUMERIC", 20, -1, false, false)
               },
               null
               );

    DDL[1] = new Table("ANNOTATION",
               new Field[]{
                  new Field("DOCID", "INTEGER", -1, -1, true, true),
                  new Field("SETNAME", "VARCHAR", 20, -1, true, false),
                  new Field("ID", "INTEGER", -1, -1, true, true),
                  new Field("TYPE", "VARCHAR", 20, -1, false, false),
                  new Field("STARTNODEID", "NUMERIC", 20, -1, true, false),
                  new Field("ENDNODEID", "NUMERIC", 20, -1, true, false)
               },
               null
             );

    DDL[2] = new Table("NODE",
               new Field[]{
                  new Field("ID", "INTEGER", -1, -1, true, true),
                  new Field("NODEOFFSET", "NUMERIC", 20, -1, true, false)
               },
               null
             );

  }
}
