/*
 *  Field.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 * 
 *  Valentin Tablan, 06/03/2000
 *
 *  $Id: Field.java,v 1.8 2000/11/08 16:34:56 hamish Exp $
 */

package gate.db;

import java.util.*;

/**
  * This class is used in defining databases structure.
  * More precisely, this class defines a column in a database table.
  */
public class Field {

  /** Debug flag */
  private static final boolean DEBUG = false;

  /**
    * Constructs a instance of Field.
    *
    * @param name a value of type 'String', the name of the column.
    * @param type a value of type 'String', the SQL type of the column.
    * @param size a value of type 'int', the size (precision) of the
    * column
    * @param scale a value of type 'int', the scale (number of decimal places)
    * of the column.
    * @param notNull a value of type 'boolean', is this column NOT NULLABLE?
    * @param primaryKey a value of type 'boolean', is this column a member
    * of the primary key?
    */
  public Field(String name, String type, int size, int scale,
               boolean notNull, boolean primaryKey){
    this.name = name;
    this.type = type;
    this.size = size;
    this.scale = scale;
    this.notNull = notNull;
    this.primaryKey = primaryKey;
  } // Field

  /**
    * Returns a String representation of this object.
    *
    * @return a value of type 'String'
    */
  public String toString(){
    String res ="'" + name +"' '"+ type + "'("+size+","+scale+") ";
    if(notNull) res += "NOT NULL ";
    if(primaryKey) res += "PRIMARY KEY";
    return res;
  } // toString

  /**
    * Checks this object for "equality" whith annother object of type Field.
    * NOTE: this is NOT an equality test. This method return true even if the
    * 'size' field of the given object is larger than the 'size' field of this
    * object. This was necessary because this method is actually used while
    * testing a database for compliance with a given structure (obviously a
    * larger size does not stop table from being compliant with the requested
    * structure).
    *
    * @param obj a value of type 'Object'
    * @return a value of type 'boolean'
    */
  public boolean equals(Object obj){
    // according to java documentation in case a contains call on a collection
    // the equals method is called on the checked object against each object
    // from that collection.
    // This means that in our case the equals method will be called on the ddl
    // column description against the actual column description found in the
    // database.
    // Out.println("Equals!");
    if(obj instanceof Field){
      boolean res = true;
      Field fieldObj = (Field)obj;

      if(name.equalsIgnoreCase(fieldObj.name));
      else return false;

      if(type.equalsIgnoreCase(fieldObj.type));
      // it seems that the numeric types get transformed depending of the
      // database implementation (e.g. for Oracle all numeric types are NUMBER)
      // and the getMetaData method reports them as DECIMAL
      else if(NUMERIC_TYPES.contains(type)&&
              NUMERIC_TYPES.contains(fieldObj.type));
      // PostgreSQL uses OID for BLOB and reports it as INTEGER
      else if(type.equalsIgnoreCase("BLOB")&&
              fieldObj.type.equalsIgnoreCase("INTEGER"));
      else if(type.equalsIgnoreCase("OTHER")||
              fieldObj.type.equalsIgnoreCase("OTHER"));
      else return false;

      // this is how it should be (we should say OK if the actual size is
      // larger than the requested one).
      if (size == -1 || fieldObj.size == -1 || size <= fieldObj.size);
      else return false;

      if(notNull == fieldObj.notNull);
      else return false;
      //PostgreSQL seems to have a problem reporting primary keys. They are
      //converted in primary indices and do not show in the getPrimaryKeys()
      //result set.
      //      if(primaryKey == fieldObj.primaryKey);
      //      else return false;

      return res;
    } else {
      return false;
    }
  } // equals

  public String name;

  public String type;

  public int size;

  public int scale;

  public boolean notNull;

  public boolean primaryKey;

  protected static Collection NUMERIC_TYPES;

  static {
    NUMERIC_TYPES = new HashSet();
    NUMERIC_TYPES.add("NUMERIC");
    NUMERIC_TYPES.add("NUMBER");
    NUMERIC_TYPES.add("INTEGER");
    NUMERIC_TYPES.add("INT");
    NUMERIC_TYPES.add("BIGINT");
    NUMERIC_TYPES.add("DOUBLE");
    NUMERIC_TYPES.add("LONG");
    NUMERIC_TYPES.add("FLOAT");
    NUMERIC_TYPES.add("DECIMAL");
    NUMERIC_TYPES.add("SMALLINT");
  }
}

