/*
 *  OracleDataStore.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 *
 *  Valentin Tablan 21 Feb 2000
 *
 *  $Id: OracleDataStore.java,v 1.9 2000/11/08 16:34:56 hamish Exp $
 */

package gate.db;

import java.sql.*;
import gate.*;
import gate.util.*;
import java.util.*;
import java.io.*;

/** Provides connection to an Oracle Database.
  */
public class OracleDataStore extends Database{

  /** Debug flag */
  private static final boolean DEBUG = false;

  /**
    * Constructs a OracleDataStore
    *
    * @param driver the driver used to connect to the database.
    * We didn't choose to default this value to any specific Oracle Jdbc
    * drivers because this option should belong to the user.
    * @param url The URL of the database.
    * @param user the username
    * @param password the password
    * @param catalog the catalog in which the Gate2 table reside.
    * @param schema The schema in which the Gate2 table reside.
    */
  public OracleDataStore(Driver driver, String url, String user,
                         String password, String catalog, String schema){
    super(driver, url, user, password, catalog, schema);
    dictionary.put("DOUBLE", "DOUBLE PRECISION");
    dictionary.put("LONG", "INTEGER");
  }

  /** Reads a Object from a BLOB on the current row in a result set.
    * Note that the Blob must contain a java Object saved using serialisation.
    * @param rs a result set positioned on the row from which the BLOB is to
    * be read
    * @param columnName the name of the column containing the BLOB on the
    * current row.
    * @return an object containing the blob.
    */
  public Object readObjectBlob(ResultSet rs,Connection con, String columnName)
                throws SQLException{
    Object ret=null;
    Blob objBlob = rs.getBlob(columnName);
    ObjectInputStream ois=null;
    InputStream is=null;
    if(objBlob !=null){

      try {
        is = objBlob.getBinaryStream();
        if(is != null) ois = new ObjectInputStream(is);
      } catch(StreamCorruptedException sce) {
        sce.printStackTrace(Err.getPrintWriter());
      } catch(IOException ioe) {
        ioe.printStackTrace(Err.getPrintWriter());
      };

      if(ois != null){
        try {
          ret = ois.readObject();
          is.close();
          ois.close();
        } catch(ClassNotFoundException cnfe) {
          cnfe.printStackTrace(Err.getPrintWriter());
        } catch(IOException ioe) {
          ioe.printStackTrace(Err.getPrintWriter());
        };
      };// if (ois != null)
    } else {
      ret = null;
    } // if (featuresBlob !=null)
    return ret;

  ////////////////////// OLD CODE ////////////////////////
  //don't feel like throwing it away yet
  /*
        //the generic JDBC way ... doesn't seem to work with ORACLE
        ObjectInputStream ois=null;
        try{
          InputStream is = rsCorpora.getBinaryStream("FEATURES");
          if(is != null){
            ois = new ObjectInputStream(is);
          }else{
            ois = null;
          }
        }catch(StreamCorruptedException sce){
          sce.printStackTrace(Err.getPrintWriter());
        }catch(IOException ioe){
          ioe.printStackTrace(Err.getPrintWriter());
        };
        if(ois != null){
          try{
            features = (FeatureMap)ois.readObject();
          }catch(ClassNotFoundException cnfe){
            cnfe.printStackTrace(Err.getPrintWriter());
          }catch(IOException ioe){
            ioe.printStackTrace(Err.getPrintWriter());
          };
        is.close();
        ois.close();
        };//if (ois != null)
*/
/*
        //the ORACLE way
        oracle.sql.BLOB featuresBlob =
            ((oracle.jdbc.driver.OracleResultSet)rsCorpora).getBLOB("FEATURES");
        ObjectInputStream ois=null;
        InputStream is=null;
        if(featuresBlob !=null){
          try{
            is = featuresBlob.getBinaryStream();
            if(is != null) ois = new ObjectInputStream(is);
          }catch(StreamCorruptedException sce){
            sce.printStackTrace(Err.getPrintWriter());
          }catch(IOException ioe){
            ioe.printStackTrace(Err.getPrintWriter());
          };
          if(ois != null){
            try{
              features = (FeatureMap)ois.readObject();
              is.close();
              ois.close();
            }catch(ClassNotFoundException cnfe){
              cnfe.printStackTrace(Err.getPrintWriter());
            }catch(IOException ioe){
              ioe.printStackTrace(Err.getPrintWriter());
            };
          };//if (ois != null)
        }else{
          features = null;
        }//if(featuresBlob !=null)
  */
  }


  /**
    * Returns a Blob representation of any serializable java object.
    *
    * @param stmt a value of type 'Statement'
    * @param obj a value of type 'Object'
    * @return a value of type 'Blob'
    * @exception SQLException if an error occurs
    */
  private Blob buildBlobFromObject(Connection con, Object obj)
              throws SQLException
  {
    if(obj == null) return null;
    String blobName = Tools.genTime().toString();
    PreparedStatement ps = con.prepareStatement(
                       "INSERT INTO FOO (NAME, TEMPBLOB) "+
                       "VALUES('"+blobName+"', empty_blob())");
    ps.executeUpdate();
    Statement stmt = con.createStatement();
    ResultSet tempRs = stmt.executeQuery("SELECT NAME, TEMPBLOB FROM FOO "+
                                         "WHERE NAME='"+blobName+"'");
    tempRs.next();
    oracle.sql.BLOB objBlob =
          ((oracle.jdbc.driver.OracleResultSet)tempRs).getBLOB("TEMPBLOB");
    OutputStream os = objBlob.getBinaryOutputStream();

    try {
      ObjectOutputStream ous = new ObjectOutputStream(os);
      ous.writeObject(obj);
      ous.close();
    } catch(IOException ioe) {
      ioe.printStackTrace(Err.getPrintWriter());
    }

    return objBlob;
  } // buildBlobFromObject

  /**
    * Acts as a setBlob for the provided prepared statement on the position
    * indicated by place.
    * @param ps A prepared statement where the blob will be set.
    * @param place The position on the prepared statement where the blob
    * will be set.
    * @param value The object that will be serialised and converted into a blob.
    */
  public void putObject(PreparedStatement ps, int place, Object value)
    throws SQLException
  {
    ps.setBlob(place, buildBlobFromObject(ps.getConnection(),value));
  } // putObject


  public Map getDictionary(){
    return dictionary;
  } // getDictionary

} // OracleDataStore


