/*
 *  TestDB.java
 *
 *  Copyright (c) 1998-2001, The University of Sheffield.
 *
 *  This file is part of GATE (see http://gate.ac.uk/), and is free
 *  software, licenced under the GNU Library General Public License,
 *  Version 2, June 1991 (in the distribution as file licence.html,
 *  and also available at http://gate.ac.uk/gate/licence.html).
 *
 *  Valentin Tablan, 23/Feb/2000
 *
 *  $Id: TestDB.java,v 1.30 2000/11/08 16:34:56 hamish Exp $
 */

package gate.db;

import junit.framework.*;
import oracle.jdbc.driver.OracleDriver;
import postgresql.Driver;

import java.io.*;
import java.sql.SQLException;
import java.util.*;
import java.net.*;

import gate.util.*;
import gate.*;
import gate.corpora.*;

public class TestDB extends TestCase
{
  /** Debug flag */
  private static final boolean DEBUG = false;

  /** Construction */
  public TestDB(String name) {
    super(name);
  }

  /** Connect to the DataStoreImpl */
  public void setUp() throws SQLException{
    ods = new OracleDataStore(new oracle.jdbc.driver.OracleDriver(),
//                        "jdbc:oracle:thin:@bradfield:1521:one",
//                        "system", "manager", null,  null);

//                        "jdbc:oracle:thin:@redmires:1521:gate",
//                        "gate", "gate", "gate",  null);
                        "jdbc:oracle:thin:@derwent:1521:dbgate",
                        "gate", "gate", "gate",  null);


    pgds = new PostgresDataStore(new postgresql.Driver(),
                      "jdbc:postgresql://redmires/gate",
                      "gate", "gate", null,null);
  } // setUp


  /** Close the db connections */
  public void tearDown() throws OpenTransactionsException{
      ods.release();
      pgds.release();
  } // tearDown

  /** Checks whether the provided database complies with Gate2 requirements */
  public void doTestDatabase(Database db){
    boolean value;

    //Check whether Gate2 is happy about the database
    value = Checker.setupDS(db);
    assert("Gate2: "+Checker.errMsg,value);

    //Check whether Corpus is happy about the database
    value = gate.db.CorpusWrapper.setupDatabase(db);
    assert("Corpus: "+Checker.errMsg,value);

    //Check whether Document is happy about the database
    value = gate.db.DocumentWrapper.setupDatabase(db);
    assert("Document: "+Checker.errMsg,value);

  } // doTestDatabase

  /**
    * Creates a new transient corpus;
    * Creates a new transient document and adds it to the newcorpus;
    * Creates a persistent corpus starting from the transient one;
    * Closes the database connection;
    * Re-opens the database connection;
    * Retrieves the previuosly created corpus and checks its contents.
    */
  public void doTestCorpus(Database db)
              throws IOException, SQLException, GateException,
                     OpenTransactionsException
  {
    //Create a Corpus
    String corpusName = "FooCorpus" + gate.util.Tools.genTime();
    Corpus someTransientCorpus = Factory.newCorpus(corpusName);

    //Add some features
    FeatureMap corpusFeatures = Factory.newFeatureMap();

    //A String value
    corpusFeatures.put("Name", corpusName);

    //A BLOB value
    Date corpusCreationTime = new Date();
    corpusFeatures.put("Time",corpusCreationTime);

    //An Integer value
    corpusFeatures.put("No of Users", new Double(10.11));

    //A Long value
    Long corpusSize = Tools.genTime();
    corpusFeatures.put("Size",corpusSize);
    someTransientCorpus.setFeatures(corpusFeatures);

    //Create a document
    Document someTransientDoc = Factory.newDocument(
             new URL("http://www.dcs.shef.ac.uk/~valyt/index.html"));

    //Add some features
    FeatureMap documentFeatures = Factory.newFeatureMap();

    //A String value
    documentFeatures.put("Author", "Johh Writer");

    //A BLOB value
    Date documentCreationTime = new Date();
    documentFeatures.put("Time",documentCreationTime);

    //An Integer value
    documentFeatures.put("No of lines", new Integer(1111));

    //A Long value
    Long documentSize = Tools.genTime();
    documentFeatures.put("Size",documentSize);
    someTransientDoc.setFeatures(documentFeatures);

    //Insert the new document in the corpus
    someTransientCorpus.add(someTransientDoc);

    //Make the corpus persistent
    CorpusWrapper persistentCorpus =
                  new CorpusWrapper(db,(CorpusImpl)someTransientCorpus);

    //Close the database
    persistentCorpus.commit();
    persistentCorpus.disconnect();
    db.release();

    //re-open the database
    if(db instanceof OracleDataStore) {
      db = new OracleDataStore(new oracle.jdbc.driver.OracleDriver(),
                       // "jdbc:oracle:thin:@bradfield:1521:one",
                       // "system", "manager", null,  null);

                       // "jdbc:oracle:thin:@redmires:1521:gate",
                       // "gate", "gate", "gate",  null);

                        "jdbc:oracle:thin:@derwent:1521:dbgate",
                        "gate", "gate", "gate",  null);

    } else if ( db instanceof PostgresDataStore ){
      db = new PostgresDataStore(new postgresql.Driver(),
                        "jdbc:postgresql://redmires/gate",
                        "gate", "gate", null,null);
    } else fail("Ooops, unknown datastore type!!");

    //read the corpus from database
    persistentCorpus = new CorpusWrapper(db,corpusName);

    //Check the features
    //=============================
    FeatureMap readFeatures = persistentCorpus.getFeatures();
    assertNotNull(readFeatures);

    String readString = (String)readFeatures.get("Name");
    assertNotNull(readString);
    assertEquals(corpusName, readString);

    Date readDate = (Date)readFeatures.get("Time");
    assertNotNull(readDate);
    assertEquals(corpusCreationTime, readDate);

    Double readDouble= (Double)readFeatures.get("No of Users");
    assertNotNull(readDouble);
    assertEquals(new Double(10.11), readDouble);

    Long readLong = (Long)readFeatures.get("Size");
    assertNotNull(readLong);
    assertEquals(corpusSize, readLong);
    //Check the document
    //=============================
    assert("The corpus is empty!",!persistentCorpus.isEmpty());
    assertEquals(1, persistentCorpus.size());
    Iterator docsIter = persistentCorpus.iterator();

    //no need for while, we've just checked there's only one element
    if(docsIter.hasNext()) {
      Document readDoc = (Document)docsIter.next();
      String actualString="<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">";

      //What we have read
      readString = readDoc.getContent().toString().substring(0,49);
      assertEquals(actualString, readString);

      readFeatures = readDoc.getFeatures();
      assertNotNull(readFeatures);

      readString = (String)readFeatures.get("Author");
      assertNotNull(readString);
      assertEquals("Johh Writer", readString);

      readDate = (Date)readFeatures.get("Time");
      assertNotNull(readDate);
      assertEquals(documentCreationTime, readDate);

      Integer readInt = (Integer)readFeatures.get("No of lines");
      assertNotNull(readInt);
      assertEquals(new Integer(1111), readInt);

      readLong = (Long)readFeatures.get("Size");
      assertNotNull(readLong);
      assertEquals(documentSize, readLong);
    }
    persistentCorpus.commit();
    persistentCorpus.disconnect();
  } // doTestCorpus

  /** Trying to read an existent corpus*/
  /*
  public void testReadCorpus() throws SQLException,
                                      GateException
  {
    //this should load my web page :)
    CorpusWrapper dbCorpus = new CorpusWrapper(db,"Test");
    //What is actually there
    String actual="<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">";
    //What we have read
    String read =
           dbCorpus.getDocument(1001).getContent().toString().substring(0,49);
    dbCorpus.commit();
    dbCorpus.disconnect();
    assertEquals(actual, read);
  } // testReadCorpus
  */

  /** tests the functionality of gate.util.ObjectWriter */
  public void testObjectWriter() throws IOException, ClassNotFoundException {

    String actual ="The quick brown fox jumps over the lazy dog!";
    ObjectWriter ow = new ObjectWriter(actual);
    ow.start();
    InputStream is = ow.getInputStream();
    String readStr=null;
    ObjectInputStream ios = new ObjectInputStream(is);
    readStr = (String) ios.readObject();
    assertEquals(actual,readStr);

    //now let's try a really large one
    actual ="";
    for(int i = 0; i<=200; i++)actual+="0123456789\n0123456789\n0123456789\n"+
                                        "0123456789\n0123456789\n0123456789\n";
    ow = new ObjectWriter(actual);
    ow.start();
    is = ow.getInputStream();
    readStr=null;
    ios = new ObjectInputStream(is);
    readStr = (String) ios.readObject();
    assertEquals(actual,readStr);
  } // testObjectWriter


  public void testDatabase() {
    //Oracle
    doTestDatabase(ods);

    //Postgresql
    doTestDatabase(pgds);
  } // testDatabase

  public void testCorpus()
              throws IOException, SQLException, GateException,
                     OpenTransactionsException
  {
    //Oracle
    doTestCorpus(ods);

    //Postgreqsl
    doTestCorpus(pgds);
  } // testCorpus

  /** Test suite routine for the test runner */
  public static Test suite() {
    return new TestSuite(TestDB.class);
  } // suite


  public static void main(String[] args) {
    try {

    TestDB testDB = new TestDB("");

    testDB.setUp();
    testDB.testDatabase();
    testDB.tearDown();

    testDB.testObjectWriter();

    testDB.setUp();
    testDB.testCorpus();
    testDB.tearDown();

    /*
    testDB.setUp();
    testDB.testReadCorpus();
    testDB.tearDown();
    testDB.setUp();
    testDB.testNewCorpus();
    */

    } catch(Exception e) {
      e.printStackTrace(Err.getPrintWriter());
    }
  }

  private OracleDataStore ods;

  private PostgresDataStore pgds;

} // TestDB
